import logging


def get_logger(name: str, level):
    # Configure the logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Remove all existing handlers
    if logger.hasHandlers():
        logger.handlers.clear()

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(filename)s - %(funcName)s - %(lineno)d - %(message)s'))
    logger.addHandler(handler)

    # Disable propagation to the root logger
    logger.propagate = False
    return logger


#!/usr/bin/env python3
from typing import Tuple, Optional, Dict

import pygame
from pygame.locals import QUIT, KEYDOWN, K_ESCAPE, K_SPACE

from airunner.api import API
from airunner.windows.pygame_window import PygameWindow
from airunner.windows.pygame_window import PygameManager
from airunner.enums import SignalCode
from airunner.handlers.llm.llm_response import LLMResponse
from airunner.handlers.llm.llm_request import LLMRequest

from airunner.pygame_example.pygame_agent import PygameAgent


class ExampleGame(PygameManager):
    """
    This simple example class demonstrates how to create a Pygame window
    using airunner, make an LLM request, and handle the response from the model
    """
    def _handle_llm_response(self, response: LLMResponse):
        print(response.message)

    def _start(self):
        self.set_screen_color()
    
    def _process_events(self):
        for event in pygame.event.get():
            if event.type == QUIT:
                self.running = False
            elif event.type == KEYDOWN and event.key == K_ESCAPE:
                self.running = False
            elif event.type == KEYDOWN and event.key == K_SPACE:
                self.api.send_llm_request("Tell me a joke.")
    
    def run(self):
        self.api.logger.info("Running pygame")
        self.running = True
        n = 0
        while self.running:
            try:
                self._process_events()
                pygame.time.delay(100)
            except Exception as e:
                self.api.logger.error(f"Error in run loop: {e}")
                self.running = False  # Ensure the loop breaks on error
    
    def set_screen_color(self, color: Optional[Tuple] = None):
        """
        Set the screen color for the Pygame window.
        """
        color = color or self.screen_color
        self.screen.fill(color)
    
    def quit(self):
        self.api.logger.info("Quitting pygame")
        self.running = False
        pygame.quit()


if __name__ == "__main__":
    airunner_api = API(
        main_window_class=PygameWindow,
        window_class_params={
            "width": 800,
            "height": 600,
            "game_class": ExampleGame,
            "local_agent_class": PygameAgent
        }
    )

    airunner_api.quit()
    pygame.quit()


from airunner.handlers.llm.agent.agents import LocalAgent


class PygameAgent(LocalAgent):
    """
    This is an example of a Pygame agent that can be used with the Pygame window.
    It inherits from the LocalAgent class and implements the system_prompt property.
    The system_prompt is used to provide the agent with a specific instruction set
    for interacting with the Pygame window.
    """
    @property
    def system_prompt(self) -> str:
        return """You are a game AI that can interact with the Pygame window."""


from abc import ABC, abstractmethod
from typing import Tuple, Type, Optional
import threading

import pygame

from PySide6.QtWidgets import QMainWindow
from PySide6.QtWidgets import QWidget
from PySide6.QtCore import QTimer, Qt
from PySide6.QtGui import QResizeEvent
from PySide6.QtWidgets import QVBoxLayout

from airunner.enums import SignalCode, LLMActionType
from airunner.workers import (
    AudioCaptureWorker,
    AudioProcessorWorker,
    LLMGenerateWorker,
    MaskGeneratorWorker,
    SDWorker,
    TTSGeneratorWorker,
    TTSVocalizerWorker,
)
from airunner.handlers.llm.llm_request import LLMRequest
from airunner.utils import create_worker
from airunner.windows.main.ai_model_mixin import AIModelMixin
from airunner.windows.main.pipeline_mixin import PipelineMixin
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.styles_mixin import StylesMixin
from airunner.mediator_mixin import MediatorMixin
from airunner.handlers.llm.llm_response import LLMResponse
from airunner.api import API
from airunner.handlers.llm.agent.agents import LocalAgent


class PygameManager(ABC):
    """
    A class to manage the Pygame window and its initialization.
    This class is responsible for setting up the Pygame environment,
    creating the window, and handling events.
    It is designed to be used with the AI Runner framework.
    This class is a wrapper around the Pygame library to provide
    a consistent interface for creating and managing a Pygame window.
    """
    def __init__(
        self, 
        api: API,
        game_title: str = "Pygame Window",
        width: int = 800,
        height: int = 600,
        screen_color: Tuple = (0, 0, 0)
    ):
        self.api = api
        self.game_title: str = game_title
        self.width: int = width
        self.height: int = height
        self.screen: pygame.Surface = None
        self.screen_color: Tuple = screen_color
        self._initialize()
        self._start()
        self.api.register(SignalCode.LLM_TEXT_STREAMED_SIGNAL, self._handle_llm_response_signal)
        self.api.logger.info("PygameManager initialized")

    def _initialize(self):
        self.api.logger.info("Initializing Pygame")
        self._initialize_pygame()
        self._initialize_screen()
        self._initialize_display()
    
    def _handle_llm_response_signal(self, data: dict):
        response = data.get("response")
        self._handle_llm_response(response)
    
    @abstractmethod
    def _handle_llm_response(self, response: LLMResponse):
        """
        Handle the LLM response.
        This method should be overridden by subclasses to provide
        specific functionality for handling LLM responses.
        """
        
    
    @abstractmethod
    def _start(self):
        """
        Start the Pygame loop.
        This method should be overridden by subclasses to provide
        specific functionality for the game loop.
        """
    
    @abstractmethod
    def run(self):
        """
        Run the Pygame loop.
        This method should be overridden by subclasses to provide
        specific functionality for the game loop.
        """
    
    @abstractmethod
    def quit(self):
        """
        Quit the Pygame loop.
        This method should be overridden by subclasses to provide
        specific functionality for quitting the game loop.
        """
        
    def _initialize_pygame(self):
        """
        Initialize the pygame library.
        """
        pygame.init()
    
    def _initialize_screen(self):
        """
        Initialize the screen for the pygame window.
        """
        self.screen = pygame.display.set_mode((self.width, self.height))
    
    def _initialize_display(self):
        """
        Initialize the display for the pygame window.
        """
        pygame.display.set_caption(self.game_title)


class PygameWidget(QWidget):
    def __init__(self, pygame_manager, parent=None):
        super().__init__(parent)
        self.pygame_manager = pygame_manager
        self.setAttribute(Qt.WidgetAttribute.WA_OpaquePaintEvent)
        self.setAttribute(Qt.WidgetAttribute.WA_PaintOnScreen)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        # Timer to update the Pygame display
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_pygame)
        self.timer.start(16)  # ~60 FPS

    def update_pygame(self):
        """
        Update the Pygame display and repaint the widget.
        """
        QTimer.singleShot(0, self._render_pygame) # MODIFIED
        self.update()

    def _render_pygame(self):
        """
        Renders the pygame scene. Must be run on the main thread.
        """
        self.pygame_manager.set_screen_color()
        pygame.display.flip()

    def paintEvent(self, event):
        """
        Override the paint event to draw the Pygame surface onto the widget.
        """
        pass  # Pygame handles its own rendering

    def resizeEvent(self, event: QResizeEvent):
        """
        Handle resizing of the widget and update the Pygame display size.
        """
        size = event.size()
        self.pygame_manager.screen = pygame.display.set_mode((size.width(), size.height()))
        super().resizeEvent(event)

    def closeEvent(self, event):
        """
        Handle the widget close event and quit Pygame.
        """
        self.pygame_manager.quit()
        pygame.quit()
        super().closeEvent(event)


class PygameWindow(
    MediatorMixin,
    SettingsMixin,
    StylesMixin,
    PipelineMixin,
    AIModelMixin,
    QMainWindow
):
    """
    A class to create a Pygame window using PySide6."

    This class is a subclass of QMainWindow and is designed to be used
    with the AI Runner framework. It initializes a Pygame window and
    manages the Pygame event loop. The class also provides methods for
    sending requests and handling signals.
    """
    def __init__(
        self, 
        app, 
        game_class: Type[PygameManager],
        width: int = 800,
        height: int = 600,
        local_agent_class: Optional[Type[LocalAgent]] = None,
        *args, 
        **kwargs
    ):
        """
        Initialize the Pygame window.
        :param app: The main application instance.
        :param game_class: The class that manages the Pygame window.
        :param width: The width of the Pygame window.
        :param height: The height of the Pygame window.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        self.app = app
        self.pygame_manager = game_class(
            api=app,
            width=width,
            height=height,
        )
        self._mask_generator_worker = create_worker(MaskGeneratorWorker)
        self._sd_worker = create_worker(SDWorker)
        self._stt_audio_capture_worker = create_worker(AudioCaptureWorker)
        self._stt_audio_processor_worker = create_worker(AudioProcessorWorker)
        self._tts_generator_worker = create_worker(TTSGeneratorWorker)
        self._tts_vocalizer_worker = create_worker(TTSVocalizerWorker)
        self._llm_generate_worker = create_worker(LLMGenerateWorker, local_agent_class=local_agent_class)

        super().__init__(*args, **kwargs)
        self.setWindowTitle("AI Runner - Pygame Window")

        # Create a central widget and layout
        central_widget = QWidget(self)
        layout = QVBoxLayout(central_widget)

        # Create and add the PygameWidget
        self.pygame_widget = PygameWidget(self.pygame_manager, self)
        layout.addWidget(self.pygame_widget)

        self.setCentralWidget(central_widget)

        # Create and start a thread for the Pygame loop
        pygame_thread = threading.Thread(target=self.pygame_manager.run)
        pygame_thread.daemon = True  # Allow the main program to exit even if the thread is still running
        pygame_thread.start()



from typing import Optional

from airunner.app import App
from airunner.handlers.llm.llm_request import LLMRequest
from airunner.enums import SignalCode, LLMActionType


class API(App):    
    def send_llm_request(self, prompt: str, llm_request: Optional[LLMRequest] = None):
        self.emit_signal(
            SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL,
            {
                "llm_request": True,
                "request_data": {
                    "action": LLMActionType.CHAT,
                    "prompt": prompt,
                    "llm_request": llm_request or LLMRequest.from_default()
                }
            }
        )


from airunner.windows.download_wizard.download_thread import DownloadThread
from airunner.windows.download_wizard.download_wizard_page import DownloadWizardPage
from airunner.windows.setup_wizard.model_setup.tts.templates.speech_t5_ui import Ui_speecht5_setup


class TTSSpeechT5Setup(DownloadWizardPage):
    class_name_ = Ui_speecht5_setup

    def start_download(self):
        self.models_to_download = [
            {
                "model": {
                    "path": self.speech_t5_settings.embeddings_path,
                    "repo_type": "dataset",
                },
            },
            {
                "model": {
                    "path": self.speech_t5_settings.vocoder_path
                },
            },
            {
                "model": {
                    "path": self.speech_t5_settings.model_path
                },
            }
        ]
        self.download_thread = DownloadThread(self.models_to_download)
        self.download_thread.progress_updated.connect(self.update_progress)
        self.download_thread.download_finished.connect(self.download_finished)
        self.download_thread.start()




import json

from typing import Callable, Dict, Optional

from airunner.enums import SignalCode
from airunner.signal_mediator import SignalMediator
from airunner.messaging.backends import RabbitMQBackend
from airunner.settings import AIRUNNER_MESSAGE_BACKEND


class MediatorMixin:
    """
    Use with any class that needs to emit and receive signals.
    Initialize with a SignalMediator instance.
    """
    _signal_handlers: Dict = {}

    def __init__(
        self, 
        mediator: Optional[SignalMediator] = None, 
        message_backend: Optional[Dict] = None,
        *args, 
        **kwargs
    ):
        """
        Initialize the mixin with an optional SignalMediator instance.
        :param mediator: Custom SignalMediator instance (e.g., with RabbitMQ backend).
        """
        if type(mediator) is not SignalMediator:
            mediator = None
        
        message_backend = message_backend or json.loads(
            AIRUNNER_MESSAGE_BACKEND or "{}"
        )

        if not mediator:
            backend = None
            if message_backend:
                backend_name = message_backend.pop("type", None)
                
                available_backends = {
                    "rabbitmq": RabbitMQBackend
                }

                if backend_name in available_backends:
                    available_backends[backend_name](**message_backend)

            mediator = SignalMediator(backend=backend)

        self.mediator = mediator

        self.register_signals()
        super().__init__(*args, **kwargs)
    
    @property
    def signal_handlers(self) -> Dict:
        return self._signal_handlers
    
    @signal_handlers.setter
    def signal_handlers(self, value):
        self._signal_handlers = value
    
    def register_signals(self):
        """
        Set signal_handlers Dict in order to register signals.

        signal_handlers should be a dictionary of SignalCode enums and functions.
        Example:
        signal_handlers = {
            SignalCode.GET_SETTINGS: self.get_settings,
            SignalCode.SET_SETTINGS: self.set_settings
        }
        :return:
        """
        for signal, handler in self.signal_handlers.items():
            self.register(signal, handler)

    def emit_signal(self, code: SignalCode, data: object = None):
        self.mediator.emit_signal(code, data)

    def register(self, code: SignalCode, slot_function: Callable):
        """
        Accessor method for SignalMediator.register method.
        :param code:
        :param slot_function:
        :return:
        """
        self.mediator.register(code, slot_function)
